# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from . import image


class Tile:
    def __init__(self, data):
        self.buffer = data

    def to_image(self):
        i = image.Image(8, 8)
        self.composite(i, 0, 0)
        return i

    def composite(self, i, x, y):
        for iy in range(8):
            ffi.memmove(ffi.addressof(i.buffer, x + (iy + y) * i.stride), ffi.addressof(self.buffer, iy * 8), 8 * ffi.sizeof("color_t"))


class CacheSet:
    def __init__(self, core):
        self.core = core
        self.cache = ffi.gc(ffi.new("struct mCacheSet*"), core._deinit_cache)
        core._init_cache(self.cache)


class TileView:
    def __init__(self, cache):
        self.cache = cache

    def get_tile(self, tile, palette):
        return Tile(lib.mTileCacheGetTile(self.cache, tile, palette))


class MapView:
    def __init__(self, cache):
        self.cache = cache

    @property
    def width(self):
        return 1 << lib.mMapCacheSystemInfoGetTilesWide(self.cache.sysConfig)

    @property
    def height(self):
        return 1 << lib.mMapCacheSystemInfoGetTilesHigh(self.cache.sysConfig)

    @property
    def image(self):
        i = image.Image(self.width * 8, self.height * 8, alpha=True)
        for y in range(self.height * 8):
            if not y & 7:
                lib.mMapCacheCleanRow(self.cache, y >> 3)
            row = lib.mMapCacheGetRow(self.cache, y)
            ffi.memmove(ffi.addressof(i.buffer, i.stride * y), row, self.width * 8 * ffi.sizeof("color_t"))
        return i


class Sprite(object):
    def constitute(self, tileView, tilePitch):
        i = image.Image(self.width, self.height, alpha=True)
        tileId = self.tile
        for y in range(self.height // 8):
            for x in range(self.width // 8):
                tile = tileView.getTile(tileId, self.paletteId)
                tile.composite(i, x * 8, y * 8)
                tileId += 1
            if tilePitch:
                tileId += tilePitch - self.width // 8
        self.image = i



# Copyright (c) 2013-2017 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from .sm83 import SM83Core
from .core import Core, needs_reset
from .memory import Memory
from .tile import Sprite
from . import create_callback


class GB(Core):
    KEY_A = lib.GBA_KEY_A
    KEY_B = lib.GBA_KEY_B
    KEY_SELECT = lib.GBA_KEY_SELECT
    KEY_START = lib.GBA_KEY_START
    KEY_DOWN = lib.GBA_KEY_DOWN
    KEY_UP = lib.GBA_KEY_UP
    KEY_LEFT = lib.GBA_KEY_LEFT
    KEY_RIGHT = lib.GBA_KEY_RIGHT

    def __init__(self, native):
        super(GB, self).__init__(native)
        self._native = ffi.cast("struct GB*", native.board)
        self.sprites = GBObjs(self)
        self.cpu = SM83Core(self._core.cpu)
        self.memory = None
        self._link = None

    @needs_reset
    def _init_cache(self, cache):
        lib.GBVideoCacheInit(cache)
        lib.GBVideoCacheAssociate(cache, ffi.addressof(self._native.video))

    def _deinit_cache(self, cache):
        lib.mCacheSetDeinit(cache)
        if self._was_reset:
            self._native.video.renderer.cache = ffi.NULL

    def _load(self):
        super(GB, self)._load()
        self.memory = GBMemory(self._core)

    def attach_sio(self, link):
        self._link = link
        lib.GBSIOSetDriver(ffi.addressof(self._native.sio), link._native)

    def __del__(self):
        if self._link:
            lib.GBSIOSetDriver(ffi.addressof(self._native.sio), ffi.NULL)
            self._link = None


create_callback("GBSIOPythonDriver", "init")
create_callback("GBSIOPythonDriver", "deinit")
create_callback("GBSIOPythonDriver", "writeSB")
create_callback("GBSIOPythonDriver", "writeSC")


class GBSIODriver(object):
    def __init__(self):
        self._handle = ffi.new_handle(self)
        self._native = ffi.gc(lib.GBSIOPythonDriverCreate(self._handle), lib.free)

    def init(self):
        return True

    def deinit(self):
        pass

    def write_sb(self, value):
        pass

    def write_sc(self, value):
        return value


class GBSIOSimpleDriver(GBSIODriver):
    def __init__(self, period=0x100):
        super(GBSIOSimpleDriver, self).__init__()
        self.rx = 0x00  # pylint: disable=invalid-name
        self._period = period

    def init(self):
        self._native.p.period = self._period
        return True

    def write_sb(self, value):
        self.rx = value  # pylint: disable=invalid-name

    def write_sc(self, value):
        self._native.p.period = self._period
        if value & 0x80:
            lib.mTimingDeschedule(ffi.addressof(self._native.p.p.timing), ffi.addressof(self._native.p.event))
            lib.mTimingSchedule(ffi.addressof(self._native.p.p.timing), ffi.addressof(self._native.p.event), self._native.p.period)
        return value

    def is_ready(self):
        return not self._native.p.remainingBits

    @property
    def tx(self):  # pylint: disable=invalid-name
        return self._native.p.pendingSB

    @property
    def period(self):
        return self._native.p.period

    @tx.setter
    def tx(self, newTx):  # pylint: disable=invalid-name
        self._native.p.pendingSB = newTx
        self._native.p.remainingBits = 8

    @period.setter
    def period(self, new_period):
        self._period = new_period
        if self._native.p:
            self._native.p.period = new_period


class GBMemory(Memory):
    def __init__(self, core):
        super(GBMemory, self).__init__(core, 0x10000)

        self.cart = Memory(core, lib.GB_SIZE_CART_BANK0 * 2, lib.GB_BASE_CART_BANK0)
        self.vram = Memory(core, lib.GB_SIZE_VRAM, lib.GB_BASE_VRAM)
        self.sram = Memory(core, lib.GB_SIZE_EXTERNAL_RAM, lib.GB_REGION_EXTERNAL_RAM)
        self.iwram = Memory(core, lib.GB_SIZE_WORKING_RAM_BANK0, lib.GB_BASE_WORKING_RAM_BANK0)
        self.oam = Memory(core, lib.GB_SIZE_OAM, lib.GB_BASE_OAM)
        self.io = Memory(core, lib.GB_SIZE_IO, lib.GB_BASE_IO)  # pylint: disable=invalid-name
        self.hram = Memory(core, lib.GB_SIZE_HRAM, lib.GB_BASE_HRAM)


class GBSprite(Sprite):
    PALETTE_BASE = (8,)

    def __init__(self, obj, core):
        self.x = obj.x  # pylint: disable=invalid-name
        self.y = obj.y  # pylint: disable=invalid-name
        self.tile = obj.tile
        self._attr = obj.attr
        self.width = 8
        lcdc = core._native.memory.io[0x40]
        self.height = 16 if lcdc & 4 else 8
        if core._native.model >= lib.GB_MODEL_CGB:
            if self._attr & 8:
                self.tile += 512
            self.palette_id = self._attr & 7
        else:
            self.palette_id = (self._attr >> 4) & 1
        self.palette_id += 8


class GBObjs:
    def __init__(self, core):
        self._core = core
        self._obj = core._native.video.oam.obj

    def __len__(self):
        return 40

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError()
        sprite = GBSprite(self._obj[index], self._core)
        sprite.constitute(self._core.tiles[0], 0)
        return sprite



from mgba.image import Image
from collections import namedtuple
from . import VideoFrame

Output = namedtuple('Output', ['video'])


class Tracer(object):
    def __init__(self, core):
        self.core = core
        self._video_fifo = []

    def yield_frames(self, skip=0, limit=None):
        self.framebuffer = Image(*self.core.desired_video_dimensions())
        self.core.set_video_buffer(self.framebuffer)
        self.core.reset()
        skip = (skip or 0) + 1
        while skip > 0:
            frame = self.core.frame_counter
            self.framebuffer = Image(*self.core.desired_video_dimensions())
            self.core.set_video_buffer(self.framebuffer)
            self.core.run_frame()
            skip -= 1
        while frame <= self.core.frame_counter and limit != 0:
            self._video_fifo.append(VideoFrame(self.framebuffer.to_pil()))
            yield frame
            frame = self.core.frame_counter
            self.core.run_frame()
            if limit is not None:
                assert limit >= 0
                limit -= 1

    def video(self, generator=None, **kwargs):
        if not generator:
            generator = self.yield_frames(**kwargs)
        try:
            while True:
                if self._video_fifo:
                    yield self._video_fifo.pop(0)
                else:
                    next(generator)
        except StopIteration:
            return



# Copyright (c) 2013-2018 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module


class Buffer(object):
    def __init__(self, native, internal_rate):
        self._native = native
        self._internal_rate = internal_rate

    @property
    def available(self):
        return lib.blip_samples_avail(self._native)

    def set_rate(self, rate):
        lib.blip_set_rates(self._native, self._internal_rate, rate)

    def read(self, samples):
        buffer = ffi.new("short[%i]" % samples)
        count = self.read_into(buffer, samples, 1, 0)
        return buffer[:count]

    def read_into(self, buffer, samples, channels=1, interleave=0):
        return lib.blip_read_samples(self._native, ffi.addressof(buffer, interleave), samples, channels == 2)

    def clear(self):
        lib.blip_clear(self._native)


class StereoBuffer(object):
    def __init__(self, left, right):
        self._left = left
        self._right = right

    @property
    def available(self):
        return min(self._left.available, self._right.available)

    def set_rate(self, rate):
        self._left.set_rate(rate)
        self._right.set_rate(rate)

    def read(self, samples):
        buffer = ffi.new("short[%i]" % (2 * samples))
        count = self.read_into(buffer, samples)
        return buffer[0:2 * count]

    def read_into(self, buffer, samples):
        samples = self._left.read_into(buffer, samples, 2, 0)
        return self._right.read_into(buffer, samples, 2, 1)

    def clear(self):
        self._left.clear()
        self._right.clear()



import numpy as np
import PIL.Image
import sys

X_FLIP = 0x4000
Y_FLIP = 0x8000
PAL = 0x1000

im = PIL.Image.open(sys.argv[1])
tiles = np.array(np.split(np.array(np.split(np.array(im), 28, 0)), 32, 2))
tilemapRaw = bytearray(32 * 28 * 2)
tilemap = np.ndarray(shape=(28, 32), dtype='<u2', buffer=tilemapRaw)
tiledata = [[0] * 64]

for tid in np.mgrid[:32, :28].reshape(2, -1).T:
    tile = tiles[tuple(tid)]
    tid = tuple(np.flip(tid))
    data = list(tile.flatten())
    try:
        index = tiledata.index(data)
        tilemap[tid] = index | PAL
        continue
    except:
        pass

    data = list(np.flip(tile, 0).flatten())
    try:
        index = tiledata.index(data)
        tilemap[tid] = index | PAL | Y_FLIP
        continue
    except:
        pass

    data = list(np.flip(tile, 1).flatten())
    try:
        index = tiledata.index(data)
        tilemap[tid] = index | PAL | X_FLIP
        continue
    except:
        pass

    data = list(np.flip(np.flip(tile, 0), 1).flatten())
    try:
        index = tiledata.index(data)
        tilemap[tid] = index | PAL | X_FLIP | Y_FLIP
        continue
    except:
        pass

    data = list(tile.flatten())
    index = len(tiledata)
    tilemap[tid] = index | PAL
    tiledata.append(data)

chardataRaw = bytearray(len(tiledata) * 32)

palette = np.array(im.getpalette(), np.uint16).reshape(-1, 3)
palette >>= 3
palette = palette[:,0] | (palette[:,1] << 5) | (palette[:,2] << 10)

for i, tile in enumerate(tiledata):
    for y in range(8):
        tileData = [0] * 4
        for x in range(8):
            t = tile[y * 8 + x]
            tileData[0] |= ((t & 1) >> 0) << (7 - x)
            tileData[1] |= ((t & 2) >> 1) << (7 - x)
            tileData[2] |= ((t & 4) >> 2) << (7 - x)
            tileData[3] |= ((t & 8) >> 3) << (7 - x)
        tileBase = i * 32 + y * 2
        chardataRaw[tileBase + 0x00] = tileData[0]
        chardataRaw[tileBase + 0x01] = tileData[1]
        chardataRaw[tileBase + 0x10] = tileData[2]
        chardataRaw[tileBase + 0x11] = tileData[3]

def printMem(mem):
    for i in range(len(mem) // 16):
        line = mem[i * 16:(i + 1) * 16]
        print('\t' + ', '.join([f'0x{b:02X}' for b in line]) + ',')

print('uint16_t palette[] = {');
for rgb in palette[:16]:
    print(f'\t0x{rgb:04X},')
print('};')
print()

print('uint8_t tilemap[] = {');
printMem(tilemapRaw)
print('};')
print()

print('uint8_t chardata[] = {');
printMem(chardataRaw)
print('};')



# Copyright (c) 2013-2017 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from .core import IRunner, ICoreOwner, Core


class ThreadCoreOwner(ICoreOwner):
    def __init__(self, thread):
        self.thread = thread

    def claim(self):
        if not self.thread.isRunning():
            raise ValueError
        lib.mCoreThreadInterrupt(self.thread._native)
        return self.thread._core

    def release(self):
        lib.mCoreThreadContinue(self.thread._native)


class Thread(IRunner):
    def __init__(self, native=None):
        if native:
            self._native = native
            self._core = Core(native.core)
            self._core._was_reset = lib.mCoreThreadHasStarted(self._native)
        else:
            self._native = ffi.new("struct mCoreThread*")

    def start(self, core):
        if lib.mCoreThreadHasStarted(self._native):
            raise ValueError
        self._core = core
        self._native.core = core._core
        lib.mCoreThreadStart(self._native)
        self._core._was_reset = lib.mCoreThreadHasStarted(self._native)

    def end(self):
        if not lib.mCoreThreadHasStarted(self._native):
            raise ValueError
        lib.mCoreThreadEnd(self._native)
        lib.mCoreThreadJoin(self._native)

    def pause(self):
        lib.mCoreThreadPause(self._native)

    def unpause(self):
        lib.mCoreThreadUnpause(self._native)

    @property
    def running(self):
        return bool(lib.mCoreThreadIsActive(self._native))

    @property
    def paused(self):
        return bool(lib.mCoreThreadIsPaused(self._native))

    def use_core(self):
        return ThreadCoreOwner(self)



#!/usr/bin/env python
from __future__ import print_function
import argparse
import errno
import os
import re
import shutil
import subprocess

qtPath = None
verbose = False

def splitPath(path):
	folders = []
	while True:
		path, folder = os.path.split(path)
		if folder != '':
			folders.append(folder)
		else:
			if path != '':
				folders.append(path)
			break
	folders.reverse()
	return folders

def joinPath(path):
	return reduce(os.path.join, path, '')

def findFramework(path):
	child = []
	while path and not path[-1].endswith('.framework'):
		child.append(path.pop())
	child.reverse()
	return path, child

def findQtPath(path):
	parent, child = findFramework(splitPath(path))
	return joinPath(parent[:-2])

def makedirs(path):
	split = splitPath(path)
	accum = []
	split.reverse()
	while split:
		accum.append(split.pop())
		newPath = joinPath(accum)
		if newPath == '/':
			continue
		try:
			os.mkdir(newPath)
		except OSError as e:
			if e.errno != errno.EEXIST:
				raise


def parseOtoolLine(line, execPath, root):
	if not line.startswith('\t'):
		return None, None, None, None
	line = line[1:]
	match = re.match(r'(\S.*) \(compatibility version.*\)', line)
	path = match.group(1)
	split = splitPath(path)
	newExecPath = ['@executable_path', '..', 'Frameworks']
	newPath = execPath[:-1]
	newPath.append('Frameworks')
	if split[:3] == ['/', 'usr', 'lib'] or split[:2] == ['/', 'System']:
		return None, None, None, None
	if split[0] == '@executable_path':
		split[:1] = execPath
	if split[0] == '/' and not os.access(joinPath(split), os.F_OK):
		split[:1] = root
	oldPath = os.path.realpath(joinPath(split))
	split = splitPath(oldPath)
	isFramework = False
	if not split[-1].endswith('.dylib'):
		isFramework = True
		split, framework = findFramework(split)
	newPath.append(split[-1])
	newExecPath.append(split[-1])
	if isFramework:
		newPath.extend(framework)
		newExecPath.extend(framework)
		split.extend(framework)
	newPath = joinPath(newPath)
	newExecPath = joinPath(newExecPath)
	return joinPath(split), newPath, path, newExecPath

def updateMachO(bin, execPath, root):
	global qtPath
	otoolOutput = subprocess.check_output([otool, '-L', bin])
	toUpdate = []
	for line in otoolOutput.split('\n'):
		oldPath, newPath, oldExecPath, newExecPath = parseOtoolLine(line, execPath, root)
		if not newPath:
			continue
		if os.access(newPath, os.F_OK):
			if verbose:
				print('Skipping copying {}, already done.'.format(oldPath))
			newPath = None
		elif os.path.abspath(oldPath) != os.path.abspath(newPath):
			if verbose:
				print('Copying {} to {}...'.format(oldPath, newPath))
			parent, child = os.path.split(newPath)
			makedirs(parent)
			shutil.copy2(oldPath, newPath)
			os.chmod(newPath, 0o644)
		toUpdate.append((newPath, oldExecPath, newExecPath))
		if not qtPath and 'Qt' in oldPath:
			qtPath = findQtPath(oldPath)
			if verbose:
				print('Found Qt path at {}.'.format(qtPath))
	args = [installNameTool]
	for path, oldExecPath, newExecPath in toUpdate:
		if path != bin:
			if path:
				updateMachO(path, execPath, root)
			if verbose:
				print('Updating Mach-O load from {} to {}...'.format(oldExecPath, newExecPath))
			args.extend(['-change', oldExecPath, newExecPath])
		else:
			if verbose:
				print('Updating Mach-O id from {} to {}...'.format(oldExecPath, newExecPath))
			args.extend(['-id', newExecPath])
	args.append(bin)
	subprocess.check_call(args)

if __name__ == '__main__':
	parser = argparse.ArgumentParser()
	parser.add_argument('-R', '--root', metavar='ROOT', default='/', help='root directory to search')
	parser.add_argument('-I', '--install-name-tool', metavar='INSTALL_NAME_TOOL', default='install_name_tool', help='path to install_name_tool')
	parser.add_argument('-O', '--otool', metavar='OTOOL', default='otool', help='path to otool')
	parser.add_argument('-p', '--qt-plugins', metavar='PLUGINS', default='', help='Qt plugins to include (comma-separated)')
	parser.add_argument('-s', '--sign', metavar='IDENTITY', help='sign with a given identity')
	parser.add_argument('-E', '--entitlements', metavar='ENTITLEMENTS', help='use a given file for entitlements when signing')
	parser.add_argument('-v', '--verbose', action='store_true', default=False, help='output more information')
	parser.add_argument('bundle', help='application bundle to deploy')
	args = parser.parse_args()

	otool = args.otool
	installNameTool = args.install_name_tool
	verbose = args.verbose

	try:
		shutil.rmtree(os.path.join(args.bundle, 'Contents/Frameworks/'))
	except OSError as e:
		if e.errno != errno.ENOENT:
			raise

	for executable in os.listdir(os.path.join(args.bundle, 'Contents/MacOS')):
		if executable.endswith('.dSYM'):
			continue
		fullPath = os.path.join(args.bundle, 'Contents/MacOS/', executable)
		updateMachO(fullPath, splitPath(os.path.join(args.bundle, 'Contents/MacOS')), splitPath(args.root))
	if args.qt_plugins:
		try:
			shutil.rmtree(os.path.join(args.bundle, 'Contents/PlugIns/'))
		except OSError as e:
			if e.errno != errno.ENOENT:
				raise
		makedirs(os.path.join(args.bundle, 'Contents/PlugIns'))
		makedirs(os.path.join(args.bundle, 'Contents/Resources'))
		with open(os.path.join(args.bundle, 'Contents/Resources/qt.conf'), 'w') as conf:
			conf.write('[Paths]\nPlugins = PlugIns\n')
		plugins = args.qt_plugins.split(',')
		for plugin in plugins:
			plugin = plugin.strip()
			kind, plug = os.path.split(plugin)
			newDir = os.path.join(args.bundle, 'Contents/PlugIns/', kind)
			makedirs(newDir)
			newPath = os.path.join(newDir, plug)
			shutil.copy2(os.path.join(qtPath, 'plugins', plugin), newPath)
			updateMachO(newPath, splitPath(os.path.join(args.bundle, 'Contents/MacOS')), splitPath(args.root))
	if args.sign:
		args = ['codesign', '-s', args.sign, '-vf', '-o', 'runtime']
		if args.entitlements:
			args.extend(['--entitlements', args.entitlements])
		args.append(args.bundle)
		subprocess.check_call(args)



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=invalid-name,unused-argument
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
import os


@ffi.def_extern()
def _vfpClose(vf):
    vfp = ffi.cast("struct VFilePy*", vf)
    ffi.from_handle(vfp.fileobj).close()
    return True


@ffi.def_extern()
def _vfpSeek(vf, offset, whence):
    vfp = ffi.cast("struct VFilePy*", vf)
    f = ffi.from_handle(vfp.fileobj)
    f.seek(offset, whence)
    return f.tell()


@ffi.def_extern()
def _vfpRead(vf, buffer, size):
    vfp = ffi.cast("struct VFilePy*", vf)
    pybuf = ffi.buffer(buffer, size)
    ffi.from_handle(vfp.fileobj).readinto(pybuf)
    return size


@ffi.def_extern()
def _vfpWrite(vf, buffer, size):
    vfp = ffi.cast("struct VFilePy*", vf)
    pybuf = ffi.buffer(buffer, size)
    ffi.from_handle(vfp.fileobj).write(pybuf)
    return size


@ffi.def_extern()
def _vfpMap(vf, size, flags):
    pass


@ffi.def_extern()
def _vfpUnmap(vf, memory, size):
    pass


@ffi.def_extern()
def _vfpTruncate(vf, size):
    vfp = ffi.cast("struct VFilePy*", vf)
    ffi.from_handle(vfp.fileobj).truncate(size)


@ffi.def_extern()
def _vfpSize(vf):
    vfp = ffi.cast("struct VFilePy*", vf)
    f = ffi.from_handle(vfp.fileobj)
    pos = f.tell()
    f.seek(0, os.SEEK_END)
    size = f.tell()
    f.seek(pos, os.SEEK_SET)
    return size


@ffi.def_extern()
def _vfpSync(vf, buffer, size):
    vfp = ffi.cast("struct VFilePy*", vf)
    f = ffi.from_handle(vfp.fileobj)
    if buffer and size:
        pos = f.tell()
        f.seek(0, os.SEEK_SET)
        res = _vfpWrite(vf, buffer, size)
        f.seek(pos, os.SEEK_SET)
        return res == size
    f.flush()
    os.fsync()
    return True


def open(f):  # pylint: disable=redefined-builtin
    handle = ffi.new_handle(f)
    vf = VFile(lib.VFileFromPython(handle), _no_gc=(f, handle))
    return vf


def open_path(path, mode="r"):
    flags = 0
    if mode.startswith("r"):
        flags |= os.O_RDONLY
    elif mode.startswith("w"):
        flags |= os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    elif mode.startswith("a"):
        flags |= os.O_WRONLY | os.O_CREAT | os.O_APPEND
    else:
        return None

    if "+" in mode[1:]:
        flags |= os.O_RDWR
    if "x" in mode[1:]:
        flags |= os.O_EXCL

    vf = lib.VFileOpen(path.encode("UTF-8"), flags)
    if vf == ffi.NULL:
        return None
    return VFile(vf)


class VFile:
    def __init__(self, vf, _no_gc=None):
        self.handle = vf
        self._no_gc = _no_gc
        self._claimed = False

    @staticmethod
    def fromEmpty():
        return VFile(lib.VFileMemChunk(ffi.NULL, 0))

    def __del__(self):
        if not self._claimed:
            self.close()

    def close(self):
        if self._claimed:
            return False
        self._claimed = True
        return bool(self.handle.close(self.handle))

    def seek(self, offset, whence):
        return self.handle.seek(self.handle, offset, whence)

    def read(self, buffer, size):
        return self.handle.read(self.handle, buffer, size)

    def read_all(self, size=0):
        if not size:
            size = self.size()
        buffer = ffi.new("char[%i]" % size)
        size = self.handle.read(self.handle, buffer, size)
        return ffi.unpack(buffer, size)

    def readline(self, buffer, size):
        return self.handle.readline(self.handle, buffer, size)

    def write(self, buffer, size):
        return self.handle.write(self.handle, buffer, size)

    def map(self, size, flags):
        return self.handle.map(self.handle, size, flags)

    def unmap(self, memory, size):
        self.handle.unmap(self.handle, memory, size)

    def truncate(self, size):
        self.handle.truncate(self.handle, size)

    def size(self):
        return self.handle.size(self.handle)

    def sync(self, buffer, size):
        return self.handle.sync(self.handle, buffer, size)



def dict_merge(a, b):
    for key, value in b.items():
        if isinstance(value, dict):
            if key in a:
                dict_merge(a[key], value)
            else:
                a[key] = dict(value)
        else:
            a[key] = value



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from .arm import ARMCore
from .core import Core, needs_reset
from .tile import Sprite
from .memory import Memory
from . import create_callback


class GBA(Core):
    KEY_A = lib.GBA_KEY_A
    KEY_B = lib.GBA_KEY_B
    KEY_SELECT = lib.GBA_KEY_SELECT
    KEY_START = lib.GBA_KEY_START
    KEY_DOWN = lib.GBA_KEY_DOWN
    KEY_UP = lib.GBA_KEY_UP
    KEY_LEFT = lib.GBA_KEY_LEFT
    KEY_RIGHT = lib.GBA_KEY_RIGHT
    KEY_L = lib.GBA_KEY_L
    KEY_R = lib.GBA_KEY_R

    SIO_NORMAL_8 = lib.SIO_NORMAL_8
    SIO_NORMAL_32 = lib.SIO_NORMAL_32
    SIO_MULTI = lib.SIO_MULTI
    SIO_UART = lib.SIO_UART
    SIO_JOYBUS = lib.SIO_JOYBUS
    SIO_GPIO = lib.SIO_GPIO

    def __init__(self, native):
        super(GBA, self).__init__(native)
        self._native = ffi.cast("struct GBA*", native.board)
        self.sprites = GBAObjs(self)
        self.cpu = ARMCore(self._core.cpu)
        self.memory = None
        self._sio = set()

    @needs_reset
    def _init_cache(self, cache):
        lib.GBAVideoCacheInit(cache)
        lib.GBAVideoCacheAssociate(cache, ffi.addressof(self._native.video))

    def _deinit_cache(self, cache):
        lib.mCacheSetDeinit(cache)
        if self._was_reset:
            self._native.video.renderer.cache = ffi.NULL

    def _load(self):
        super(GBA, self)._load()
        self.memory = GBAMemory(self._core, self._native.memory.romSize)

    def attach_sio(self, link, mode=lib.SIO_MULTI):
        self._sio.add(mode)
        lib.GBASIOSetDriver(ffi.addressof(self._native.sio), link._native, mode)

    def __del__(self):
        for mode in self._sio:
            lib.GBASIOSetDriver(ffi.addressof(self._native.sio), ffi.NULL, mode)


create_callback("GBASIOPythonDriver", "init")
create_callback("GBASIOPythonDriver", "deinit")
create_callback("GBASIOPythonDriver", "load")
create_callback("GBASIOPythonDriver", "unload")
create_callback("GBASIOPythonDriver", "writeRegister")


class GBASIODriver(object):
    def __init__(self):
        super(GBASIODriver, self).__init__()

        self._handle = ffi.new_handle(self)
        self._native = ffi.gc(lib.GBASIOPythonDriverCreate(self._handle), lib.free)

    def init(self):
        return True

    def deinit(self):
        pass

    def load(self):
        return True

    def unload(self):
        return True

    def write_register(self, address, value):
        return value


class GBASIOJOYDriver(GBASIODriver):
    RESET = lib.JOY_RESET
    POLL = lib.JOY_POLL
    TRANS = lib.JOY_TRANS
    RECV = lib.JOY_RECV

    def __init__(self):
        super(GBASIOJOYDriver, self).__init__()

        self._native = ffi.gc(lib.GBASIOJOYPythonDriverCreate(self._handle), lib.free)

    def send_command(self, cmd, data):
        buffer = ffi.new('uint8_t[5]')
        try:
            buffer[0] = data[0]
            buffer[1] = data[1]
            buffer[2] = data[2]
            buffer[3] = data[3]
            buffer[4] = data[4]
        except IndexError:
            pass

        outlen = lib.GBASIOJOYSendCommand(self._native, cmd, buffer)
        if outlen > 0 and outlen <= 5:
            return bytes(buffer[0:outlen])
        return None


class GBAMemory(Memory):
    def __init__(self, core, romSize=lib.SIZE_CART0):
        super(GBAMemory, self).__init__(core, 0x100000000)

        self.bios = Memory(core, lib.SIZE_BIOS, lib.BASE_BIOS)
        self.wram = Memory(core, lib.SIZE_WORKING_RAM, lib.BASE_WORKING_RAM)
        self.iwram = Memory(core, lib.SIZE_WORKING_IRAM, lib.BASE_WORKING_IRAM)
        self.io = Memory(core, lib.SIZE_IO, lib.BASE_IO)  # pylint: disable=invalid-name
        self.palette = Memory(core, lib.SIZE_PALETTE_RAM, lib.BASE_PALETTE_RAM)
        self.vram = Memory(core, lib.SIZE_VRAM, lib.BASE_VRAM)
        self.oam = Memory(core, lib.SIZE_OAM, lib.BASE_OAM)
        self.cart0 = Memory(core, romSize, lib.BASE_CART0)
        self.cart1 = Memory(core, romSize, lib.BASE_CART1)
        self.cart2 = Memory(core, romSize, lib.BASE_CART2)
        self.cart = self.cart0
        self.rom = self.cart0
        self.sram = Memory(core, lib.SIZE_CART_SRAM, lib.BASE_CART_SRAM)


class GBASprite(Sprite):
    TILE_BASE = 0x800, 0x400
    PALETTE_BASE = 0x10, 1

    def __init__(self, obj):
        self._a = obj.a
        self._b = obj.b
        self._c = obj.c
        self.x = self._b & 0x1FF  # pylint: disable=invalid-name
        self.y = self._a & 0xFF  # pylint: disable=invalid-name
        self._shape = self._a >> 14
        self._size = self._b >> 14
        self._256_color = bool(self._a & 0x2000)
        self.width, self.height = lib.GBAVideoObjSizes[self._shape * 4 + self._size]
        self.tile = self._c & 0x3FF
        if self._256_color:
            self.palette_id = 0
            self.tile >>= 1
        else:
            self.palette_id = self._c >> 12


class GBAObjs:
    def __init__(self, core):
        self._core = core
        self._obj = core._native.video.oam.obj

    def __len__(self):
        return 128

    def __getitem__(self, index):
        if index >= len(self):
            raise IndexError()
        sprite = GBASprite(self._obj[index])
        tiles = self._core.tiles[3 if sprite._256_color else 2]
        map_1d = bool(self._core._native.memory.io[0] & 0x40)
        sprite.constitute(tiles, 0 if map_1d else 0x20)
        return sprite



#!/usr/bin/env python
import os
import os.path
import shutil
from configparser import ConfigParser

suffixes = {
    'C': 'CGB',
    'S': 'SGB',
    'A': 'AGB',
    'mgb': 'MGB',
    'sgb': 'SGB',
    'sgb2': 'SGB2',
    'cgb': 'CGB',
    'agb': 'AGB',
    'ags': 'AGB',
}

def ingestDirectory(path, dest):
    for root, _, files in os.walk(path, topdown=False):
        root = root[len(os.path.commonprefix([root, path])):]
        if root.startswith('utils'):
            continue
        for file in files:
            fname, ext = os.path.splitext(file)
            if ext not in ('.gb', '.sym'):
                continue

            try:
                os.makedirs(os.path.join(dest, root, fname))
            except OSError:
                pass

            if ext in ('.gb', '.sym'):
                shutil.copy(os.path.join(path, root, file), os.path.join(dest, root, fname, 'test' + ext))

            for suffix, model in suffixes.items():
                if fname.endswith('-' + suffix):
                    manifest = ConfigParser()
                    try:
                        with open(os.path.join(dest, root, fname, 'config.ini'), 'r') as f:
                            manifest.read_file(f)
                    except IOError:
                        pass
                    manifest.set('ports.cinema', 'gb.model', model)
                    with open(os.path.join(dest, root, fname, 'config.ini'), 'w') as f:
                        manifest.write(f, space_around_delimiters=False)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Update mooneye-gb test suite')
    parser.add_argument('source', type=str, help='directory containing built tests')
    parser.add_argument('dest', type=str, nargs='?', default=os.path.dirname(__file__), help='directory to contain ingested tests')
    args = parser.parse_args()

    ingestDirectory(args.source, args.dest)



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi  # pylint: disable=no-name-in-module
from . import png

try:
    import PIL.Image as PImage
except ImportError:
    pass


class Image:
    def __init__(self, width, height, stride=0, alpha=False):
        self.width = width
        self.height = height
        self.stride = stride
        self.alpha = alpha
        self.constitute()

    def constitute(self):
        if self.stride <= 0:
            self.stride = self.width
        self.buffer = ffi.new("color_t[{}]".format(self.stride * self.height))

    def save_png(self, fileobj):
        png_file = png.PNG(fileobj, mode=png.MODE_RGBA if self.alpha else png.MODE_RGB)
        success = png_file.write_header(self)
        success = success and png_file.write_pixels(self)
        png_file.write_close()
        return success

    if 'PImage' in globals():
        def to_pil(self):
            colorspace = "RGBA" if self.alpha else "RGBX"
            return PImage.frombytes(colorspace, (self.width, self.height), ffi.buffer(self.buffer), "raw",
                                    colorspace, self.stride * 4)


def u16_to_u32(color):
    # pylint: disable=invalid-name
    r = color & 0x1F
    g = (color >> 5) & 0x1F
    b = (color >> 10) & 0x1F
    a = (color >> 15) & 1
    abgr = r << 3
    abgr |= g << 11
    abgr |= b << 19
    abgr |= (a * 0xFF) << 24
    return abgr


def u32_to_u16(color):
    # pylint: disable=invalid-name
    r = (color >> 3) & 0x1F
    g = (color >> 11) & 0x1F
    b = (color >> 19) & 0x1F
    a = color >> 31
    abgr = r
    abgr |= g << 5
    abgr |= b << 10
    abgr |= a << 15
    return abgr


if ffi.sizeof("color_t") == 2:
    def color_to_u16(color):
        return color

    color_to_u32 = u16_to_u32  # pylint: disable=invalid-name

    def u16_to_color(color):
        return color

    u32_to_color = u32_to_u16  # pylint: disable=invalid-name
else:
    def color_to_u32(color):
        return color

    color_to_u16 = u32_to_u16  # pylint: disable=invalid-name

    def u32_to_color(color):
        return color

    u16_to_color = u16_to_u32  # pylint: disable=invalid-name



#!/usr/bin/env python
from __future__ import print_function
import argparse
import csv
import os
import shlex
import signal
import socket
import subprocess
import sys
import time

class PerfTest(object):
    EXECUTABLE = 'mgba-perf'

    def __init__(self, rom, renderer='software'):
        self.rom = rom
        self.renderer = renderer
        self.results = None
        self.name = 'Perf Test: {}'.format(rom)

    def get_args(self):
        return []

    def wait(self, proc):
        pass

    def run(self, cwd):
        args = [os.path.join(os.getcwd(), self.EXECUTABLE), '-P']
        args.extend(self.get_args())
        if not self.renderer:
            args.append('-N')
        elif self.renderer == 'threaded-software':
            args.append('-T')
        args.append(self.rom)
        env = {}
        if 'LD_LIBRARY_PATH' in os.environ:
            env['LD_LIBRARY_PATH'] = os.path.abspath(os.environ['LD_LIBRARY_PATH'])
            env['DYLD_LIBRARY_PATH'] = env['LD_LIBRARY_PATH'] # Fake it on OS X
        proc = subprocess.Popen(args, stdout=subprocess.PIPE, cwd=cwd, universal_newlines=True, env=env)
        try:
            self.wait(proc)
            proc.wait()
        except:
            proc.kill()
            raise
        if proc.returncode:
            print('Game crashed!', file=sys.stderr)
            return
        reader = csv.DictReader(proc.stdout)
        self.results = next(reader)

class WallClockTest(PerfTest):
    def __init__(self, rom, duration, renderer='software'):
        super(WallClockTest, self).__init__(rom, renderer)
        self.duration = duration
        self.name = 'Wall-Clock Test ({} seconds, {} renderer): {}'.format(duration, renderer, rom)

    def wait(self, proc):
        time.sleep(self.duration)
        proc.send_signal(signal.SIGINT)

class GameClockTest(PerfTest):
    def __init__(self, rom, frames, renderer='software'):
        super(GameClockTest, self).__init__(rom, renderer)
        self.frames = frames
        self.name = 'Game-Clock Test ({} frames, {} renderer): {}'.format(frames, renderer, rom)

    def get_args(self):
        return ['-F', str(self.frames)]

class PerfServer(object):
    ITERATIONS_PER_INSTANCE = 50
    RETRIES = 4

    def __init__(self, address, root='/', command=None):
        s = address.rsplit(':', 1)
        if len(s) == 1:
            self.address = (s[0], 7216)
        else:
            self.address = (s[0], s[1])
        self.command = None
        if command:
            self.command = shlex.split(command)
        self.iterations = self.ITERATIONS_PER_INSTANCE
        self.socket = None
        self.results = []
        self.reader = None
        self.root = root

    def _start(self, test):
        if self.command:
            server_command = list(self.command)
        else:
            server_command = [os.path.join(os.getcwd(), PerfTest.EXECUTABLE)]
        server_command.extend(['-PD'])
        if hasattr(test, "frames"):
            server_command.extend(['-F', str(test.frames)])
        if not test.renderer:
            server_command.append('-N')
        elif test.renderer == 'threaded-software':
            server_command.append('-T')
        for backoff in range(self.RETRIES):
            try:
                subprocess.check_call(server_command)
                break
            except subprocess.CalledProcessError as e:
                print("Failed to start server:", e, file=sys.stderr)
                if backoff == self.RETRIES - 1:
                    raise
                time.sleep(2 ** backoff)
        time.sleep(3)
        for backoff in range(self.RETRIES):
            try:
                self.socket = socket.create_connection(self.address, timeout=1000)
                break
            except OSError as e:
                print("Failed to connect:", e, file=sys.stderr)
                if backoff == self.RETRIES - 1:
                    raise
                time.sleep(2 ** backoff)
        kwargs = {}
        if sys.version_info[0] >= 3:
            kwargs["encoding"] = "utf-8"
        self.reader = csv.DictReader(self.socket.makefile(**kwargs))

    def run(self, test):
        if not self.socket:
            self._start(test)
        self.socket.send(os.path.join(self.root, test.rom).encode("utf-8"))
        self.results.append(next(self.reader))
        self.iterations -= 1
        if self.iterations == 0:
            self.finish()
            self.iterations = self.ITERATIONS_PER_INSTANCE

    def finish(self):
        self.socket.send(b"\n");
        self.reader = None
        self.socket.close()
        time.sleep(5)
        self.socket = None

class Suite(object):
    def __init__(self, cwd, wall=None, game=None, renderer='software'):
        self.cwd = cwd
        self.tests = []
        self.wall = wall
        self.game = game
        self.renderer = renderer
        self.server = None

    def set_server(self, server):
        self.server = server

    def collect_tests(self):
        roms = []
        for f in os.listdir(self.cwd):
            if f.endswith('.gba') or f.endswith('.zip') or f.endswith('.gbc') or f.endswith('.gb'):
                roms.append(f)
        roms.sort()
        for rom in roms:
            self.add_tests(rom)

    def add_tests(self, rom):
        if self.wall:
            self.tests.append(WallClockTest(rom, self.wall, renderer=self.renderer))
        if self.game:
            self.tests.append(GameClockTest(rom, self.game, renderer=self.renderer))

    def run(self):
        results = []
        sock = None
        for test in self.tests:
            print('Running test {}'.format(test.name), file=sys.stderr)
            last_result = None
            if self.server:
                self.server.run(test)
                last_result = self.server.results[-1]
            else:
                try:
                    test.run(self.cwd)
                except KeyboardInterrupt:
                    print('Interrupted, returning early...', file=sys.stderr)
                    return results
                if test.results:
                    results.append(test.results)
                    last_result = results[-1]
            if last_result:
                print('{:.2f} fps'.format(int(last_result['frames']) * 1000000 / float(last_result['duration'])), file=sys.stderr)
        if self.server:
            self.server.finish()
            results.extend(self.server.results)
        return results

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-w', '--wall-time', type=float, default=0, metavar='TIME', help='wall-clock time')
    parser.add_argument('-g', '--game-frames', type=int, default=0, metavar='FRAMES', help='game-clock frames')
    parser.add_argument('-N', '--disable-renderer', action='store_const', const=True, help='disable video rendering')
    parser.add_argument('-T', '--threaded-renderer', action='store_const', const=True, help='threaded video rendering')
    parser.add_argument('-s', '--server', metavar='ADDRESS', help='run on server')
    parser.add_argument('-S', '--server-command', metavar='COMMAND', help='command to launch server')
    parser.add_argument('-o', '--out', metavar='FILE', help='output file path')
    parser.add_argument('-r', '--root', metavar='PATH', type=str, default='/perfroms', help='root path for server mode')
    parser.add_argument('directory', help='directory containing ROM files')
    args = parser.parse_args()

    renderer = 'software'
    if args.disable_renderer:
        renderer = None
    elif args.threaded_renderer:
        renderer = 'threaded-software'
    s = Suite(args.directory, wall=args.wall_time, game=args.game_frames, renderer=renderer)
    if args.server:
        if args.server_command:
            server = PerfServer(args.server, args.root, args.server_command)
        else:
            server = PerfServer(args.server, args.root)
        s.set_server(server)
    s.collect_tests()
    results = s.run()
    fout = sys.stdout
    if args.out:
        fout = open(args.out, 'w')
    writer = csv.DictWriter(fout, results[0].keys())
    writer.writeheader()
    writer.writerows(results)
    if fout is not sys.stdout:
        fout.close()



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module
from . import vfs

MODE_RGB = 0
MODE_RGBA = 1
MODE_INDEX = 2


class PNG:
    def __init__(self, f, mode=MODE_RGB):
        self._vfile = vfs.open(f)
        self._png = None
        self._info = None
        self.mode = mode

    def write_header(self, image):
        self._png = lib.PNGWriteOpen(self._vfile.handle)
        if self.mode == MODE_RGB:
            self._info = lib.PNGWriteHeader(self._png, image.width, image.height)
        if self.mode == MODE_RGBA:
            self._info = lib.PNGWriteHeaderA(self._png, image.width, image.height)
        if self.mode == MODE_INDEX:
            self._info = lib.PNGWriteHeader8(self._png, image.width, image.height)
        return self._info != ffi.NULL

    def write_pixels(self, image):
        if self.mode == MODE_RGB:
            return lib.PNGWritePixels(self._png, image.width, image.height, image.stride, image.buffer)
        if self.mode == MODE_RGBA:
            return lib.PNGWritePixelsA(self._png, image.width, image.height, image.stride, image.buffer)
        if self.mode == MODE_INDEX:
            return lib.PNGWritePixels8(self._png, image.width, image.height, image.stride, image.buffer)
        return False

    def write_close(self):
        lib.PNGWriteClose(self._png, self._info)
        self._png = None
        self._info = None



from setuptools import setup
import re
import os
import os.path
import sys
import subprocess


def get_version_component(piece):
    return subprocess.check_output(['cmake', '-DPRINT_STRING={}'.format(piece), '-P', '../../../version.cmake']).decode('utf-8').strip()


version = '{}.{}.{}'.format(*(get_version_component(p) for p in ('LIB_VERSION_MAJOR', 'LIB_VERSION_MINOR', 'LIB_VERSION_PATCH')))
if not get_version_component('GIT_TAG'):
    version += '.{}+g{}'.format(*(get_version_component(p) for p in ('GIT_REV', 'GIT_COMMIT_SHORT')))

setup(
    name="mgba",
    version=version,
    author="Jeffrey Pfau",
    author_email="jeffrey@endrift.com",
    url="http://github.com/mgba-emu/mgba/",
    packages=["mgba"],
    setup_requires=['cffi>=1.6', 'pytest-runner'],
    install_requires=['cffi>=1.6', 'cached-property'],
    extras_require={'pil': ['Pillow>=2.3'], 'cinema': ['pytest']},
    tests_require=['pytest'],
    cffi_modules=["_builder.py:ffi"],
    license="MPL 2.0",
    classifiers=[
        "Programming Language :: C",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Topic :: Games/Entertainment",
        "Topic :: System :: Emulators"
    ]
)



import os
import os.path
import mgba.core
import mgba.image
import cinema.movie
import glob
import re
from copy import deepcopy
from cinema import VideoFrame
from cinema.util import dict_merge
from configparser import ConfigParser


class CinemaTest(object):
    TEST = 'test.(mvl|gb|gba|nds)'

    def __init__(self, path, root, settings={}):
        self.full_path = path or []
        self.path = os.path.abspath(os.path.join(root, *self.full_path))
        self.root = root
        self.name = '.'.join(path)
        self.settings = settings
        try:
            with open(os.path.join(self.path, 'config.ini'), 'r') as f:
                cfg = ConfigParser()
                cfg.read_file(f)
                settings = {}
                if 'testinfo' in cfg:
                    settings = dict(cfg['testinfo'])
                if 'ports.cinema' in cfg:
                    settings['config'] = dict(cfg['ports.cinema'])
                dict_merge(self.settings, settings)
        except IOError:
            pass
        self.tests = {}

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def setup(self):
        results = [f for f in glob.glob(os.path.join(self.path, 'test.*')) if re.search(self.TEST, f)]
        self.core = mgba.core.load_path(results[0])
        if 'config' in self.settings:
            self.config = mgba.core.Config(defaults=self.settings['config'])
            self.core.load_config(self.config)
            self.core.reset()

    def add_test(self, name, cls=None, settings={}):
        cls = cls or self.__class__
        new_settings = deepcopy(self.settings)
        dict_merge(new_settings, settings)
        self.tests[name] = cls(self.full_path + [name], self.root, new_settings)
        return self.tests[name]

    def output_settings(self):
        output_settings = {}
        if 'frames' in self.settings:
            output_settings['limit'] = int(self.settings['frames'])
        if 'skip' in self.settings:
            output_settings['skip'] = int(self.settings['skip'])
        return output_settings

    def __lt__(self, other):
        return self.path < other.path


class VideoTest(CinemaTest):
    BASELINE = 'baseline_%04u.png'

    def setup(self):
        super(VideoTest, self).setup()
        self.tracer = cinema.movie.Tracer(self.core)

    def generate_frames(self):
        for i, frame in enumerate(self.tracer.video(**self.output_settings())):
            try:
                baseline = VideoFrame.load(os.path.join(self.path, self.BASELINE % i))
                yield baseline, frame, VideoFrame.diff(baseline, frame)
            except IOError:
                yield None, frame, (None, None)

    def test(self):
        self.baseline, self.frames, self.diffs = zip(*self.generate_frames())
        assert not any(any(diffs[0].image.convert("L").point(bool).getdata()) for diffs in self.diffs)

    def generate_baseline(self):
        for i, frame in enumerate(self.tracer.video(**self.output_settings())):
            frame.save(os.path.join(self.path, self.BASELINE % i))


def gather_tests(root=os.getcwd()):
    tests = CinemaTest([], root)
    for path, _, files in os.walk(root):
        test = [f for f in files if re.match(CinemaTest.TEST, f)]
        if not test:
            continue
        prefix = os.path.commonprefix([path, root])
        suffix = path[len(prefix)+1:]
        test_path = suffix.split(os.sep)
        test_root = tests
        for component in test_path[:-1]:
            new_test = test_root.tests.get(component)
            if not new_test:
                new_test = test_root.add_test(component)
            test_root = new_test
        test_root.add_test(test_path[-1], VideoTest)
    return tests



import numpy as np
import PIL.Image
import PIL.ImageChops
import sys

blocksize = 104
pow = [0] * 256
rev = [0] * 256
rev[0] = 0xFF
x = 1
for i in range(0xFF):
    pow[i]=x
    rev[x]=i
    x *= 2
    if x >= 0x100:
        x ^= 0x187
gg = [0] * 16
gg[0] = pow[0x78]
for i in range(16):
    gg[i] = 1
    for j in range(i + 1):
        j = i - j
        if j == 0:
            y = 0
        else:
            y = gg[j - 1]
        x = gg[j]
        if x != 0:
            x = rev[x] + 0x78 + i
            if x >= 0xFF:
                x -= 0xFF
            y ^= pow[x]
        gg[j] = y
for i in range(16):
    gg[i] = rev[gg[i]]

def interleave(data, header):
    data = data.reshape([-1, 48]).T
    new_data = np.zeros((64, data.shape[1]), dtype=data.dtype)
    for i, row in enumerate(data.T):
        new_data[:,i] = rs(row)
    data = new_data.reshape([-1])
    new_data = np.zeros(data.shape[0] + (102 - data.shape[0] % 102), dtype=data.dtype)
    new_data[:data.shape[0]] = data
    new_data = new_data.reshape([-1, 102])
    data = new_data
    new_data = np.zeros((data.shape[0], 104), dtype=data.dtype)
    new_data[:,2:] = data
    for i in range(new_data.shape[0]):
        x = (i * 2) % len(header)
        new_data[i,:2] = header[x:x + 2]
    data = new_data.reshape([-1])
    return data

def rs(data):
    new_data = np.zeros(data.shape[0] + 16, dtype=data.dtype)
    new_data[:data.shape[0]] = data
    new_data = np.flipud(new_data)
    for i in range(data.shape[0]):
        i = new_data.shape[0] - i - 1
        z = rev[new_data[i] ^ new_data[15]]
        for j in range(16):
            j = 15 - j
            if j == 0:
                x = 0
            else:
                x = new_data[j - 1]
            if z != 0xFF:
                y = gg[j]
                if y != 0xFF:
                    y += z
                    if y >= 0xFF:
                        y -= 0xFF
                    x ^= pow[y]
            new_data[j] = x
    new_data[:16] = ~new_data[:16]
    new_data = np.flipud(new_data)
    return new_data

def bin2raw(data):
    if len(data) == 1344:
        header = np.array([0x00, 0x02, 0x00, 0x01, 0x40, 0x10, 0x00, 0x1C], dtype=np.uint8)
    else:
        header = np.array([0x00, 0x03, 0x00, 0x19, 0x40, 0x10, 0x00, 0x2C], dtype=np.uint8)
    header = rs(header)
    new_data = interleave(np.frombuffer(data, np.uint8), header)
    return new_data.tobytes()

def make_dotcode(data):
    size = len(data)

    if size in (1344, 2112):
        data = bin2raw(data)
        size = len(data)

    blocks = size // blocksize
    height = 36
    width = 35
    margin = 2

    dots = np.zeros((width * blocks + margin * 2 + 1, height + margin * 2), dtype=bool)
    anchor = np.array([[0, 1, 1, 1, 0],
                       [1, 1, 1, 1, 1],
                       [1, 1, 1, 1, 1],
                       [1, 1, 1, 1, 1],
                       [0, 1, 1, 1, 0]], dtype=bool)
    alignment = np.array([1, 1, 1, 1, 1, 0, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0], dtype=bool)
    nybbles = [[0, 0, 0, 0, 0], [0, 0, 0, 0, 1], [0, 0, 0, 1, 0], [1, 0, 0, 1, 0],
               [0, 0, 1, 0, 0], [0, 0, 1, 0, 1], [0, 0, 1, 1, 0], [1, 0, 1, 1, 0],
               [0, 1, 0, 0, 0], [0, 1, 0, 0, 1], [0, 1, 0, 1, 0], [1, 0, 1, 0, 0],
               [0, 1, 1, 0, 0], [0, 1, 1, 0, 1], [1, 0, 0, 0, 1], [1, 0, 0, 0, 0]]
    addr = [0x03FF]
    for i in range(1, 54):
        addr.append(addr[i - 1] ^ ((i & -i) * 0x769))
        if (i & 0x07) == 0:
            addr[i] ^= 0x769
        if (i & 0x0F) == 0:
            addr[i] ^= 0x769 << 1
        if (i & 0x1F) == 0:
            addr[i] ^= (0x769 << 2) ^ 0x769

    base = 1 if blocks == 18 else 25
    for i in range(blocks + 1):
        dots[i * width:i * width + 5, 0:5] = anchor
        dots[i * width:i * width + 5, height + margin * 2 - 5:height + margin * 2] = anchor
        dots[i * width + margin, margin + 5] = 1
        a = addr[base + i]
        for j in range(16):
            dots[i * width + margin, margin + 14 + j] = a & (1 << (15 - j))
    for i in range(blocks):
        dots[i * width:(i + 1) * width, margin] = alignment
        dots[i * width:(i + 1) * width, height + margin - 1] = alignment
        block = []
        for byte in data[i * blocksize:(i + 1) * blocksize]:
            block.extend(nybbles[byte >> 4])
            block.extend(nybbles[byte & 0xF])
        j = 0
        for y in range(3):
            dots[i * width + margin + 5:i * width + margin + 31, margin + 2 + y] = block[j:j + 26]
            j += 26
        for y in range(26):
            dots[i * width + margin + 1:i * width + margin + 35, margin + 5 + y] = block[j:j + 34]
            j += 34
        for y in range(3):
            dots[i * width + margin + 5:i * width + margin + 31, margin + 31 + y] = block[j:j + 26]
            j += 26
    return np.pad(dots.T, 2)

with open(sys.argv[1], 'rb') as f:
    dots = make_dotcode(f.read())

im = PIL.Image.fromarray(dots)
im = PIL.ImageChops.invert(im)
im.save('dotcode.bmp')



# Copyright (c) 2013-2016 Jeffrey Pfau
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ._pylib import ffi, lib  # pylint: disable=no-name-in-module


class MemoryView(object):
    def __init__(self, core, width, size, base=0, sign="u"):
        self._core = core
        self._width = width
        self._size = size
        self._base = base
        self._bus_read = getattr(self._core, "busRead" + str(width * 8))
        self._bus_write = getattr(self._core, "busWrite" + str(width * 8))
        self._raw_read = getattr(self._core, "rawRead" + str(width * 8))
        self._raw_write = getattr(self._core, "rawWrite" + str(width * 8))
        self._mask = (1 << (width * 8)) - 1  # Used to force values to fit within range so that negative values work
        if sign == "u" or sign == "unsigned":
            self._type = "uint{}_t".format(width * 8)
        elif sign == "i" or sign == "s" or sign == "signed":
            self._type = "int{}_t".format(width * 8)
        else:
            raise ValueError("Invalid sign type: '{}'".format(sign))

    def _addr_check(self, address):
        if isinstance(address, slice):
            start = address.start or 0
            stop = self._size - self._width if address.stop is None else address.stop
        else:
            start = address
            stop = address + self._width
        if start >= self._size or stop > self._size:
            raise IndexError()
        if start < 0 or stop < 0:
            raise IndexError()

    def __len__(self):
        return self._size

    def __getitem__(self, address):
        self._addr_check(address)
        if isinstance(address, slice):
            start = address.start or 0
            stop = self._size - self._width if address.stop is None else address.stop
            step = address.step or self._width
            return [int(ffi.cast(self._type, self._bus_read(self._core, self._base + a))) for a in range(start, stop, step)]
        return int(ffi.cast(self._type, self._bus_read(self._core, self._base + address)))

    def __setitem__(self, address, value):
        self._addr_check(address)
        if isinstance(address, slice):
            start = address.start or 0
            stop = self._size - self._width if address.stop is None else address.stop
            step = address.step or self._width
            for addr in range(start, stop, step):
                self._bus_write(self._core, self._base + addr, value[addr] & self._mask)
        else:
            self._bus_write(self._core, self._base + address, value & self._mask)

    def raw_read(self, address, segment=-1):
        self._addr_check(address)
        return int(ffi.cast(self._type, self._raw_read(self._core, self._base + address, segment)))

    def raw_write(self, address, value, segment=-1):
        self._addr_check(address)
        self._raw_write(self._core, self._base + address, segment, value & self._mask)


class MemorySearchResult(object):
    def __init__(self, memory, result):
        self.address = result.address
        self.segment = result.segment
        self.guessDivisor = result.guessDivisor
        self.type = result.type

        if result.type == Memory.SEARCH_INT:
            if result.width == 1:
                self._memory = memory.u8
            elif result.width == 2:
                self._memory = memory.u16
            elif result.width == 4:
                self._memory = memory.u32
        elif result.type == Memory.SEARCH_STRING:
            self._memory = memory.u8
        else:
            raise ValueError("Unknown type: %X" % result.type)

    @property
    def value(self):
        if self.type == Memory.SEARCH_STRING:
            raise ValueError
        return self._memory[self.address] * self.guessDivisor

    @value.setter
    def value(self, v):
        if self.type == Memory.SEARCH_STRING:
            raise IndexError
        self._memory[self.address] = v // self.guessDivisor


class Memory(object):
    SEARCH_INT = lib.mCORE_MEMORY_SEARCH_INT
    SEARCH_STRING = lib.mCORE_MEMORY_SEARCH_STRING
    SEARCH_GUESS = lib.mCORE_MEMORY_SEARCH_GUESS

    SEARCH_EQUAL = lib.mCORE_MEMORY_SEARCH_EQUAL

    READ = lib.mCORE_MEMORY_READ
    WRITE = lib.mCORE_MEMORY_READ
    RW = lib.mCORE_MEMORY_RW

    def __init__(self, core, size, base=0):
        self.size = size
        self.base = base
        self._core = core

        self.u8 = MemoryView(core, 1, size, base, "u")
        self.u16 = MemoryView(core, 2, size, base, "u")
        self.u32 = MemoryView(core, 4, size, base, "u")
        self.s8 = MemoryView(core, 1, size, base, "s")
        self.s16 = MemoryView(core, 2, size, base, "s")
        self.s32 = MemoryView(core, 4, size, base, "s")

    def __len__(self):
        return self.size

    def search(self, value, type=SEARCH_GUESS, flags=RW, limit=10000, old_results=[]):
        results = ffi.new("struct mCoreMemorySearchResults*")
        lib.mCoreMemorySearchResultsInit(results, len(old_results))
        params = ffi.new("struct mCoreMemorySearchParams*")
        params.memoryFlags = flags
        params.type = type
        params.op = self.SEARCH_EQUAL
        if type == self.SEARCH_INT:
            params.valueInt = int(value)
        else:
            params.valueStr = ffi.new("char[]", str(value).encode("ascii"))

        for result in old_results:
            native_result = lib.mCoreMemorySearchResultsAppend(results)
            native_result.address = result.address
            native_result.segment = result.segment
            native_result.guessDivisor = result.guessDivisor
            native_result.type = result.type
        if old_results:
            lib.mCoreMemorySearchRepeat(self._core, params, results)
        else:
            lib.mCoreMemorySearch(self._core, params, results, limit)
        new_results = [MemorySearchResult(self, lib.mCoreMemorySearchResultsGetPointer(results, i)) for i in range(lib.mCoreMemorySearchResultsSize(results))]
        lib.mCoreMemorySearchResultsDeinit(results)
        return new_results

    def __getitem__(self, address):
        if isinstance(address, slice):
            return bytearray(self.u8[address])
        return self.u8[address]
